import { Connection } from "../connection/Connection";
import { QueryRunner } from "../query-runner/QueryRunner";
import { SubjectCollection } from "./subject/SubjectCollection";
/**
 * Executes PersistOperation in the given connection.
 */
export declare class PersistSubjectExecutor {
    private connection;
    private queryRunner;
    constructor(connection: Connection, queryRunner: QueryRunner);
    /**
     * Executes given persist operation.
     */
    execute(subjects: SubjectCollection): Promise<void>;
    /**
     * Executes insert operations.
     *
     * For insertion we separate two groups of entities:
     * - first group of entities are entities which do not have any relations
     *      or entities which do not have any non-nullable relation
     * - second group of entities are entities which does have non-nullable relations
     *
     * Insert process of the entities from the first group which can only have nullable relations are actually a two-step process:
     * - first we insert entities without their relations, explicitly left them NULL
     * - later we update inserted entity once again with id of the object inserted with it
     *
     * Yes, two queries are being executed, but this is by design.
     * There is no better way to solve this problem and others at the same time.
     *
     * Insert process of the entities from the second group which can have only non nullable relations is a single-step process:
     * - we simply insert all entities and get into attention all its dependencies which were inserted in the first group
     */
    private executeInsertOperations(insertSubjects);
    /**
     * Inserts an entity from the given insert operation into the database.
     * If entity has an generated column, then after saving new generated value will be stored to the InsertOperation.
     * If entity uses class-table-inheritance, then multiple inserts may by performed to save all entities.
     */
    private insert(subject, alreadyInsertedSubjects);
    private collectColumnsAndValues(metadata, entity, date, parentIdColumnValue, discriminatorValue, alreadyInsertedSubjects);
    /**
     * Executes insert operations for closure tables.
     */
    private executeInsertClosureTableOperations(insertSubjects);
    private insertClosureTableValues(subject, insertedSubjects);
    /**
     * Executes update operations.
     */
    private executeUpdateOperations(updateSubjects);
    private update(subject);
    private executeUpdateRelations(subjects);
    private updateRelations(subject);
    /**
     * Executes remove operations.
     */
    private executeRemoveOperations(removeSubjects);
    private remove(subject);
    /**
     * Executes insert junction operations.
     */
    private executeInsertJunctionsOperations(subjects, insertSubjects);
    /**
     * Executes insert junction operation.
     */
    private insertJunctions(subject, junctionInsert, insertSubjects);
    /**
     * Executes remove junction operations.
     */
    private executeRemoveJunctionsOperations(subjects);
    /**
     * Executes remove junction operation.
     */
    private removeJunctions(subject, junctionRemove);
    /**
     * Updates all special columns of the saving entities (create date, update date, versioning).
     */
    private updateSpecialColumnsInPersistedEntities(insertSubjects, updateSubjects, removeSubjects);
}
